#define _POSIX_SOURCE	1
/* 
 * config-file utility module for twd
 * dannyb@bga.com	10/27/96
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>

#include "rcfile.h"

#define TRUE	1
#define FALSE	0

/* Notes:
 * dup2array() makes malloced copies of each string.
 * free2array() frees each string then the entire array.
 * Same for dup3array() and free3array().
 * sto2array() only mallocs room for the pointers and does not copy
 * the original strings. It does trash the strings, however.
 * load_rcfile() is the main routine here. It loads a text file,
 * strips comments and '\0' chars, and parses into a char *** token 
 * array. It then makes a fully-malloced copy of the array and frees
 * the original file buffer and temporary array. The resulting array
 * can be freed with free3array() when no longer needed.
 * show3array() and show2array() are debug routines to display *** and
 * ** arrays respectively.
*/

char ***load_rcfile(char *filename) {
  char *filebuf;
  char ***linewordarray;
  static char ***result;
  int line;

  filebuf = loadtextfile(filename);
  if (filebuf == NULL) {
	return NULL;
  }
  linewordarray = sto3array(filebuf);
  if (linewordarray == NULL) {
	free(filebuf);
	return NULL;
  }
  result = dup3array(linewordarray);

  /* free the original stuff */
  for(line=0; linewordarray[line] != NULL; line ++) {
	free(linewordarray[line]);
  }
  free(linewordarray);
  free(filebuf);

  return result;
}


int statfile(char *filename) {
/*
 * compare the file's modification date with the value saved from a
 * previous call. If the file has changed, return TRUE, else
 * return FALSE. Returns -1 on error.
*/
  struct stat stats;
  static unsigned long old_date = -1;
  unsigned long new_date;
  int result;
  if (stat(filename, &stats) != 0)
	return -1;
  new_date=(long)stats.st_mtime;
  if ( (old_date != -1) && (old_date != new_date) )
	result = TRUE;
  else
	result = FALSE;
  old_date = new_date;
  return result;
}


/*****************************************************************
 * Routines for "arrays of arrays of strings".                   *
 *****************************************************************/

void free3array(char ***array) {
  int i;
  for(i=0; array[i] != NULL; i++) {
	free2array(array[i]);
  }
  free(array);
  return;
}


void show3array(char ***array) {
  int line, word;
  for(line=0; array[line]!=NULL; line++) {
	for(word=0; array[line][word]!=NULL; word++)
		printf("[%s]", array[line][word]);
	printf("\n");
  }
  return;
}



char ***dup3array(char ***array) {
  static char ***r;
  int lines = 0;
  int i;
  int j;

  for(lines=0; array[lines] != NULL; lines++);
  if (lines == 0)
	return NULL;
  r = calloc(lines+1, sizeof(char *));
  if (r == NULL) {
	return NULL;
  }

  for(i=j=0; array[i] != NULL; i++) {
		r[j] = dup2array(array[i]);
		if (r[j] == NULL) {	/* mem error */
			/* free everything malloc'ed so far */
			free3array(r);
			return NULL;
		}
		j++;
  }
  r[j] = NULL;	/* line count including NULL arg */
  if (j == 0) {	/* no matching lines */
	free(r);
	return NULL;
  }
  return r;
}


char ***sto3array(char *s) {
  int linecount;
  int i, j;
  char **l_array;
  char ***lw_array;

  l_array = sto2array(s, '\n');
  if (l_array == NULL)
	return NULL;

  for(linecount=0; l_array[linecount] != NULL; linecount++);
  lw_array = (char ***)calloc(linecount+1, sizeof(char *));
  if (lw_array == NULL) {
	free(l_array);
	return NULL;
  }
  for(i=j=0; l_array[i] != NULL; i++) {
	lw_array[j] = sto2array(l_array[i], ' ');
	if (lw_array[j] != NULL)
		j++;
  }
  lw_array[j] = NULL;
  free(l_array);
  return (char ***)realloc(lw_array, (j+1)*sizeof(char *) );
}


/********************************************************************
 * Routines for "NULL terminated arrays of strings".                *
 ********************************************************************/


void show2array(char **array) {
  printf("show2array: ");
  while(*array != NULL) {
        printf("[%s]", *array);
	array++;
  }
  printf("\n");
  return;
} 


void free2array(char **array) {
  int i;
  for(i=0; array[i] != NULL; i++) { free(array[i]); }
  free(array);
  return;
}


char **dup2array(char **array) {
/*
 * Copy an array of strings, malloc()ing for each string. Each element
 * in the array must be free()ed individually before the array can
 * be free()ed. The whole array can be free()ed by calling free2array().
 * Returns NULL if calling array is 0 strings long or on error.
 * Calls:	free2array(), count2array().
*/
  static char **result;
  int count;
  int i;

  for(count=0; array[count] != NULL; count++);
  if (count == 0)
	return NULL;

  result = calloc(count+1, sizeof(char *));
  if (result == NULL)
	return NULL;

  for(i=0; array[i] != NULL; i++) {
	result[i] = malloc(strlen(array[i]) + 1);
	if (result[i] == NULL) {
		free2array(result);	/* free entire array */
		return NULL;
	} else {
		strcpy(result[i], array[i]);
	}
  }
  result[i] = NULL;
  return result;
}


char **sto2array(char *buf, char delim) {
/*
 * Break a string(buf) consisting of a number of tokens delimited by
 * delim into an NULL-terminated array or char *pointers. Blank lines
 * are ignored. Returns NULL on error or if no tokens found.
*/ 
  static char **array;
  char *index;
  int offset;
  int instring;
  int tokencount;

  /* count the tokens in the buffer */
  tokencount = 0;
  index = buf;
  instring = FALSE;
  while(*index != '\0') {
	if ( (instring == FALSE) && (*index != delim) ) {
		/* start of a token */
		instring = TRUE;
		tokencount++;
	}
	if ( (instring == TRUE) && (*index == delim) ) {
		/* end of a token */
		instring = FALSE;
	}
	/* in the middle of a token */
	index++;
  }
  if (tokencount < 1) {
	return NULL;	/* no useful lines available */
  }
  /* malloc an array big enough to hold a pointer to each token */
  array = (char **)calloc(tokencount+1, sizeof(char *));
  if (array == NULL)
	return NULL;

  /* parse buf into strings, zap delim chars */
  index = buf;
  offset = 0;
  instring = FALSE;
  while(*index != '\0') {
	if ( (instring == FALSE) && (*index != delim) ) {
		/* start of token */
		instring = TRUE;
		array[offset] = index;
		offset++;
	}
	if ( (instring == TRUE) && (*index == delim) ) {
		/* end of token */
		*index = '\0';
		instring = FALSE;
	}
	/* middle of token */
	index++;
  }
  array[tokencount] = NULL;	/* final NULL-terminator */
  return array;
}

char *array2tos(char **array, char delim) {
/*
 * Convert an array of strings into one big string. Stuff a 'delim' char 
 * between each string.
*/
  int i;
  int tokencount;
  int bufsize;
  static char *result;
  char delimstr[2];

  if (array == NULL)
	return NULL;

  /* count strings in array */
  for(tokencount=0; array[tokencount] != NULL; tokencount++);
  if (tokencount < 1)
	return NULL;

  /* Calculate total length of strings in array */
  bufsize = 0;
  for(i=0; i<tokencount; i++) {
	bufsize += strlen(array[i]);
  }
  bufsize += tokencount;	/* add in space for delims */

  /* malloc the space */
  result = malloc((size_t)bufsize+1);
  if (result == NULL)
	return NULL;

  /* convert delim char to a string */
  delimstr[0] = delim;
  delimstr[1] = '\0';

  /* concatenate each array arg + delimiter to result string */
  result[0] = '\0';
  for(i=0; i<tokencount; i++) {
	strcat(result, array[i]);
	strcat(result, delimstr);
  }

  return result;
}


/*****************************************************************
 * Misc support routines                                         *
 *****************************************************************/


int samestring(char *s1, char *s2) {
/* A case-insensitive version of strcmp(). */

  while( (*s1 != '\0') && (*s2 != '\0') ) {
        if (toupper(*s1++) != toupper(*s2++))
                return FALSE;
  }
  return TRUE;
}


/*****************************************************************
 * Textfile routines                                             *
 *****************************************************************/

char *loadtextfile(char *filename) {
/*
 * Allocate space for contents of filename, read file into buffer
 * and return result. All '\0' chars are converted to newlines so 
 * result will be one big-ass string. Return NULL on error or if
 * file is 0 bytes. Result can be free()ed when no longer needed.
*/
  int fd;
  struct stat statbuf;
  static char *filebuf;
  int filesize;
  char *bufptr;
  fd = open(filename, O_RDONLY);
  if (fd == -1) {
	return NULL;
  }

  if (fstat(fd, &statbuf) != 0) {
	close(fd);
	return NULL;
  }
  filesize = (int)statbuf.st_size;
  if (filesize < 1) {
	close(fd);
	return NULL;
  }
  filebuf = malloc((size_t)filesize + 1);
  if (filebuf == NULL) {
	close(fd);
	return NULL;
  }

  if (readfile(fd, filebuf) < 0) {
	close(fd);
	free(filebuf);
	return NULL;
  }

  close(fd);
  bufptr = realloc(filebuf, strlen(filebuf) + 1);
  if (bufptr == NULL)
	return filebuf;
  return bufptr;
}



int readfile(int fd, char *filebuf) {
/*
 * Load open file (fd) into supplied buf. Return -1 if a read() error
 * occurs, else return 0.
*/
  int ign_comments = TRUE;	/* zap all "#" comment lines to EOL */
  int b_slash_esc = TRUE;	/* allow newlines to be escaped */
  int w_sp_to_sp = TRUE;	/* convert tabs,etc to ascii space */
  char *bufptr;
  int r;

  bufptr = filebuf;
  while(1) {
	r = read(fd, bufptr, 1);
	if (r < 0)	/* error on read() call */
		return -1;
	if (r == 0) {	/* end of file */
		bufptr = '\0';
		return 0;
	}

	/* strip out anything non-ASCII */
	if (*bufptr < 0)
		continue;

	/* convert NULs to newlines */
	if (*bufptr == '\0')
		*bufptr = '\n';

	/* convert whitespaces to ascii space char */ 
	if ( (w_sp_to_sp) && isspace(*bufptr) && (*bufptr != '\n'))
		*bufptr = ' ';

	/* ignore comments till EOL */
	if ( (ign_comments) && (*bufptr == '#') ) {	
		while(1) {
			r = read(fd, bufptr, 1);
			if (r < 0)	/* error on read() call */
				return -1;
			if (r == 0) {	/* end of file */
				bufptr = '\0';
				return 0;
			}
			if (*bufptr == '\n')
				break;
		}
	}		

	/* zap a backslash-newline pair */
	if ( (b_slash_esc) && (*bufptr == '\\') ) {
		r = read(fd, bufptr, 1);
		if (r < 0)	/* error on read() call */
			return -1;
		if (r == 0) {	/* end of file */
			bufptr = '\0';
			return 0;
		}
		if (*bufptr == '\n')
			continue;
	}
	bufptr++;
  }
  return -1;	/* never happens */
}
